/* Sistema Caotico: ondas

Para compilar:
gcc ondas.cc -lXaw -lXmu -lXext -lXt -lX11 -ldl -lm -L /usr/X11/lib */
     
#include <stdio.h>
#include <stdlib.h>            
#include <stdarg.h>         
#include <time.h> 
#include <math.h>     
#include <sys/timeb.h>

#define XK_MISCELLANY 1
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysymdef.h>

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>

#ifdef DO_SHM
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif /* DO_SHM */


// resolucao
#define RESX 400
#define RESY 400

// numero de pontos a serem ondulados
// 100
#define NUMPX 300

// amplitude inicial da onda
#define AMPL 50

// inercia
#define INE 0.0002

// velocidade de ondulacao
#define VEL 0.001

// grau de erro (1 = 100% = amplitude inicial da onda)
// 0.2
#define ERR 0

// deslocamento (se o ponto estah entre -1 e 1, y)
// 10
#define DESL 0

// valor de cada ponto
struct stPonto {
  float fY;   // posicao Y (altura)
  float fV;   // vetor deslocamento
};

// conjunto de pontos do sistema
stPonto
  astPONTOSX [ NUMPX ];


#define DEPTH 8
Display* pDpy = NULL;
XtAppContext stApContext;
Widget stTopLevel = 0, stShell = 0, stBox;

int nScreenNum;
int nDpyWidth, nDpyHeight;
Window stWindow;
XImage* pXImg = NULL;
GC stGC;

#ifdef DO_SHM
static XShmSegmentInfo stShmi;
XImage stShmImg;
#endif

unsigned char *pVid = NULL;

Colormap stColorMap = 0;

static XrmOptionDescRec table[] = {
    {"-flag",   "Flag",                XrmoptionNoArg,  (caddr_t) "on"},
    {"-param",  "parameter",           XrmoptionSepArg, NULL},
};

typedef struct {
    int flag;
    int parameter;
    char *geometry;
    int borderwidth;
} stAppResourceRec, *pAppRes;

static stAppResourceRec stAppResources;
static char szWinDimension [ 100 ];

static XtResource resources[] = {
 {"param", "Parameter", XtRInt, sizeof(int),
   XtOffset(pAppRes, parameter), XtRImmediate, (caddr_t) 80},
 {"flag", "Flag", XtRBoolean, sizeof(Boolean),
   XtOffset(pAppRes, flag), XtRImmediate, (caddr_t) TRUE },
 {XtNgeometry,  "Geometry", XtRString, sizeof(caddr_t),
   XtOffset(pAppRes, geometry), XtRString, (caddr_t) szWinDimension},
 {XtNborderWidth, "borderWidth", XtRInt, sizeof(int),
   XtOffset(pAppRes, borderwidth), XtRImmediate, (caddr_t) 1},
};



/* Atualiza area de video
**********************************************************/
void UpDate ()
{
  # ifdef DO_SHM
    XShmPutImage (pDpy, stWindow, stGC, &stShmImg, 0, 0, 0, 0, RESX, RESY, False);
  # endif
  XPutImage (pDpy, stWindow, stGC, pXImg, 0, 0, 0, 0, RESX, RESY);
}



/* Marca um pixel na area de video
**********************************************************/
void SetPixel ( int x, int y, int color )
{
  if ( x < 0 || x >= RESX ||
       y < 0 || y >= RESY )
    return;

  * ( pVid + RESX * y +x ) = color;
}



/* Draw a line with the specific color. Low optimized code.
Give the points (nXI, nYI) -> (nXF,nYF). Do line clip.
**********************************************************/
void Line ( int nXI, int nYI, int nXF, int nYF, char yCol )
{
   // define start and end Y (the less and greater Y)
   int
      nYStart = nYI,
      nYEnd = nYF,
      nSignalY = 1;
   if ( nYF < nYStart ) {
      nYStart = nYF;
      nYEnd = nYI;
      nSignalY = -1;
   }

   // define start and end X (the less and greater X)
  int
      nXStart = nXI,
      nXEnd = nXF,
      nSignalX = 1;
   if ( nXF < nXStart ) {
      nXStart = nXF;
      nXEnd = nXI;
      nSignalX = -1;
   }

   // compute delta between the given points
  int
    nDeltaX = nXF - nXI,
    nDeltaY = nYF - nYI;

   // if delta X is the greater you draw the line using Y as function of X
   if ( abs ( nDeltaX ) > abs ( nDeltaY ) ) {

      float
        fYRate = (float) nDeltaY / nDeltaX, // define Y rate
          fY;                                            // start Y
      // define start line
      if ( nSignalY == -1 )
          fY = (float) nYEnd;
     else
       fY = (float) nYStart;

     // draw line
      // test if X is growing left (1) or right (-1)
      if ( nSignalX == 1 )
         for ( int nX = nXStart; nX < nXEnd + 1; nX ++ ) {

            SetPixel ( nX, (int) fY, yCol );

            fY += fYRate;
         }
      else {

        fYRate = - fYRate;

         for ( int nX = nXEnd; nX > nXStart; nX -- ) {

            SetPixel ( nX, (int) fY, yCol );

            fY += fYRate;
         }
      }
   }
   else {   // draw the line using X as function of Y

      float
        fXRate = (float) nDeltaX / nDeltaY, // define X rate
          fX;                                            // start X
      // define start colum
      if ( nSignalX == -1 )
          fX = (float) nXEnd;
      else
          fX = (float) nXStart;

     // draw line
      // test if Y is growing down (1) or up (-1)
      if ( nSignalY == 1 )
         for ( int nY = nYStart; nY < nYEnd + 1; nY ++ ) {

            SetPixel ( (int) fX, nY, yCol );

            fX += fXRate;
         }
      else {

        fXRate = - fXRate;

         for ( int nY = nYEnd; nY > nYStart; nY -- ) {

            SetPixel ( (int) fX, nY, yCol );

            fX += fXRate;
         }
      }
   }
}



/* Limpa area de video
**********************************************************/
void Clear (int color)
{
  if ( pVid )
    memset ( pVid, color, RESX * RESY );
}



/* Atualiza janela com o grafico dos pontos da onda
**********************************************************/
void Grafico ()
{
  // limpa janela
  Clear ( 0 );

  // linha horizontal (y=0)
  Line ( 0, RESY / 2, RESX, RESY / 2, 1 );

  // faz grafico
  // espaco entre cada ponto
  int
    nEspaco = RESX / ( NUMPX + 1 );    // +1: folga no inicio e fim
  // varre (para cada ponto)
  int
    nX0 = 0,
    nY0 = 0,
    nX1,
    nY1;
  for ( int ncX = 0; ncX < NUMPX; ncX ++ ) {

    nX1 = ( ncX + 1 ) * nEspaco;
    nY1 = (int) astPONTOSX [ ncX ].fY;

    if ( ncX != 0 )
      Line ( nX0, nY0 + RESY / 2, nX1, nY1 + RESY / 2, 1 );

    nX0 = nX1;
    nY0 = nY1;
  }

  UpDate ();
}



/* Executa a animacao: faz interacoes do sistema
**********************************************************/
void AnimHandler ( XtPointer , XtIntervalId* )
{
  // altera posicao dos pontos conforme forcas
  // para cada ponto
  for ( int ncP = 0; ncP < NUMPX; ncP ++ ) {

    // distancia entre ponto
    float
      fDistSeg,  // seguinte
      fDistAnt;  // anterior
    // seguinte
    // se for o ultimo ponto calcula distancia com o primeiro
    if ( ncP == NUMPX - 1 )
      fDistSeg = astPONTOSX [ 0 ].fY - astPONTOSX [ ncP ].fY;
    else
      fDistSeg = astPONTOSX [ ncP + 1 ].fY - astPONTOSX [ ncP ].fY;
    // anterior
    // se for o primeiro ponto calcula distancia com o ultimo
    if ( ncP == 0 )
      fDistAnt = astPONTOSX [ NUMPX - 1 ].fY - astPONTOSX [ 0 ].fY;
    else
      fDistAnt = astPONTOSX [ ncP - 1 ].fY - astPONTOSX [ ncP ].fY;

    // aceleracao para o ponto (eh seguindo os pontos laterais)
    float
      fAc = ( fDistSeg + fDistAnt / 2 ) * INE;

    // forca contraria???!!! (para a outra metade), pressao?
    float
      fPr;
    if ( astPONTOSX [ ncP ].fY > 0 )
      fPr = - VEL * astPONTOSX [ ncP ].fY * astPONTOSX [ ncP ].fY;
    else
      fPr = VEL * astPONTOSX [ ncP ].fY * astPONTOSX [ ncP ].fY;

    // altera forcas
    astPONTOSX [ ncP ].fV += fAc + fPr;

    // altera posicao
    astPONTOSX [ ncP ].fY += astPONTOSX [ ncP ].fV;
  }

  // atualiza janela
  Grafico ();
    
  XtAppAddTimeOut ( stApContext, 5, AnimHandler, NULL );
}  



/* Termina janela
**********************************************************/
Term ()
{
  if ( stShell )
    XtPopdown ( stShell );
  if ( stTopLevel )
    XtDestroyWidget ( stTopLevel );
}



/* Trata eventos do X Windows
**********************************************************/
void EventsHandler (
  Widget w,
  XtPointer closure,
  XEvent* event,
  Boolean* cont_to_dispatch)
{
  KeySym key;
  int key2;
  int state;
  
  //(void)w; (void)closure; (void)cont_to_dispatch;

  switch ( event->type )
  {
    case Expose:
      if ( ! pXImg )
      {
  pXImg = XGetImage (pDpy, stWindow, 0, 0, RESX, RESY, AllPlanes, ZPixmap);
        #ifdef DO_SHM
  {
    stShmImg = *pXImg;
    stShmi.shmid = shmget (IPC_PRIVATE, pXImg->bytes_per_line*pXImg->height,
             IPC_CREAT|0777);
    stShmi.shmaddr = (char*)shmat (stShmi.shmid, 0, 0);
    stShmi.readOnly = FALSE;
    XShmAttach (pDpy, &stShmi);
    stShmImg.data = stShmi.shmaddr;
    pVid = (unsigned char*)stShmi.shmaddr;
    stShmImg.obdata = (char*)&stShmi;
  }
        #endif /* DO_SHM */
  {
    pVid = (unsigned char*)(pXImg->data);
        }
      }
      break;

    case KeyPress:
      key = XLookupKeysym ((XKeyEvent*)event, 0);
      state = ((XKeyEvent*)event)->state;
      key2 = key;
      switch (key)
      {
  case XK_Escape:
    Term ();
    exit ( 0 );
    break;
  default:
          break;
      }
      break;

    default:
      break;
  }
}



/* Inicia o aplicativo
**********************************************************/
IniApl ( int argc, char* argv [] )
{
  // cria janela, palette, define eventos, ...

  XGCValues stValues;
  Screen* pScreenPtr;
  Widget stBox;

  sprintf (szWinDimension, "%dx%d", RESX, RESY);

  stTopLevel = XtVaAppInitialize (&stApContext, "Ondas", table, XtNumber (table),
  &argc, argv, NULL, NULL, 0);

  XtGetApplicationResources (stTopLevel, &stAppResources, resources,
  XtNumber(resources), NULL, (Cardinal) 0);
  /*XParseGeometry (stAppResources.geometry, &geom_x, &geom_y, &geom_width,
  &geom_height);*/

  stShell = XtVaCreatePopupShell ("- Ondas -",
        topLevelShellWidgetClass,
        stTopLevel,
        XtNgeometry, stAppResources.geometry,
        XtNborderWidth, stAppResources.borderwidth,
        0);
  /*stBox = XtCreateManagedWidget ("box", boxWidgetClass, stShell, NULL, (Cardinal) 0);*/
  stBox = XtCreateWidget ("box", boxWidgetClass, stShell, NULL, (Cardinal) 0);

  XtPopup (stShell, XtGrabNone);
 
  pDpy = XtDisplay (stShell);
  nScreenNum = DefaultScreen (pDpy);
  pScreenPtr = DefaultScreenOfDisplay (pDpy);
  nDpyWidth = DisplayWidth (pDpy, nScreenNum);
  nDpyHeight = DisplayHeight (pDpy, nScreenNum);
  stWindow = XtWindow (stShell);

  stGC = XCreateGC (pDpy, stWindow, 0, &stValues);
  XSetForeground (pDpy, stGC, BlackPixel (pDpy, nScreenNum));
  XSetLineAttributes (pDpy, stGC, 0, LineSolid, CapButt, JoinMiter);
  XSetGraphicsExposures (pDpy, stGC, False);

  stColorMap = XCreateColormap (pDpy, RootWindow (pDpy, nScreenNum),
        DefaultVisual (pDpy, nScreenNum), AllocAll);

  XtAddEventHandler (stShell, ExposureMask | KeyPressMask | KeyReleaseMask |
           KeymapStateMask | FocusChangeMask |
         PointerMotionMask | ButtonPressMask | ButtonReleaseMask, True,
         EventsHandler, NULL);

  XMapWindow (pDpy, XtWindow (stShell));
  XSetWindowColormap (pDpy, XtWindow (stShell), stColorMap);

  pXImg = NULL;

  // mensagem de inicio
  printf ( "Sistema Caotico: ondas.\n");
  #ifdef DO_SHM
  printf ( "Usando SHM\n" );
  #endif
  #ifndef DO_SHM
    printf ( "Sem SHM\n" );
  #endif
  
  // programa palette
  XColor
    color;
  for ( int ncI = 0 ; ncI < 256; ncI ++ ) {
    color.pixel = ncI;
    color.red = 64 * 256;
    color.green = ( ( ncI * 20 ) % 256 ) * 256;
    color.blue = 255 * 256;
    color.flags = DoRed | DoGreen | DoBlue;
    XStoreColor ( pDpy, stColorMap, &color );
  }
  // fundo
  color.pixel = 0;
  color.red = 0 * 256;
  color.green = 0 * 256;
  color.blue = 0 * 256;
  color.flags = DoRed | DoGreen | DoBlue;
  XStoreColor ( pDpy, stColorMap, &color );
  // grafico
  color.pixel = 1;
  color.red = 255 * 256;
  color.green = 255 * 256;
  color.blue = 255 * 256;
  color.flags = DoRed | DoGreen | DoBlue;
  XStoreColor ( pDpy, stColorMap, &color );
}



main ( int argc, char* argv [] )
{
  // inicia matriz de pontos
  for ( int ncP = 0; ncP < NUMPX; ncP ++ ) {

    // define angulo em radianos
    float
      fRad = (float) ( 2 ) * M_PI * (float) ( ncP ) / NUMPX;

    // calcula delta (erro) - no maximo 20% da amplitude
    float
      fErro = ERR * AMPL * random () / RAND_MAX;

    // altura do ponto: seno + erro
    astPONTOSX [ ncP ].fY = sin ( fRad ) * ( AMPL / 2 ) + fErro;
    // se for entre -1 e 1, altera valor
    if ( astPONTOSX [ ncP ].fY > -1 && astPONTOSX [ ncP ].fY < 1 )
      astPONTOSX [ ncP ].fY += DESL;
    // deslocamento
    astPONTOSX [ ncP ].fV = 0;
  }

  IniApl ( argc, argv );

  XtAppAddTimeOut ( stApContext, 300, AnimHandler, NULL );
  XtAppMainLoop ( stApContext );

  Term ();
}
